########################################################
# Sanchez J. Time Series for Data Scientists, CUP 2023
# Ch10 Case study 3                                    # 
#       Using tsfeatures package                                             #
# Using Automated feature generation                   #
#  Unsupervised Machine Learning. The tsfeatures package #
# Clustering the automated features of Birmingham parks # 
######################################################

# Set working directory to source file location 

# We feature the tsfeatures package in this program.
# Install it (uncomment it if not in your library)
# install.packages("tsfeatures")
library(tsfeatures)

########### Birmingham data set program 
#########################################
park <- read.csv("dataset.csv")
attach(park)
dim(park)
   
            #############
            # STEP 1    #
            #############
head(park)
tail(park)
##### Just testing how many are negative percent or 
### beyond 100% occupancy rate -Obviously these are mistakes
occupancy=park$Occupancy; capacity=park$Capacity
test= park 
test$percent= occupancy/capacity
head(test)
badpercent=test[test$percent>1,]
dim(badpercent)
table(badpercent$SystemCodeNumber)
badmin=test[test$percent< 0,]
dim(badmin)
table(badmin$SystemCodeNumber)

## Investigate parks with less than 1312 observations. 
## for example
#park[SystemCodeNumber=="Bull Ring",]

########## Back to the original data 
table(SystemCodeNumber)
dim(park)


sum(is.na(park[, 1]))
sum(is.na(park[, 2]))
sum(is.na(park[, 3]))
sum(is.na(park[, 4]))

# There are no NA's explicitly, so we  explored the data more.

park[park$Occupancy <= 0, ]

# 1) Parking lots with SystemCodeNumber "BHMBCCPST01" and
#"BHMBRCBRG01" both had 7 0-values.
# "NIA North" had values that were less than or equal 0. While a parking
# lot can have 0 occupied spots (i.e. it's empty), having negative occupancy
# isn't logical.

# 2) Suspect dates:

park[park$SystemCodeNumber == "BHMBCCPST01" & park$Occupancy <= 0, ]

# BHMBCCPST01 had 2 "0 occupancy days" on 10/22/16
# Total: 2 0-values on 1 day.

park[park$SystemCodeNumber == "BHMBRCBRG01" & park$Occupancy <= 0, ]

# BHMBRCBRG01 had 5 0-occupancy days across 11/22/16 and 11/23/16
# Total: 5 0-values on 1 day.

park[park$SystemCodeNumber == "NIA North" & park$Occupancy <= 0, ]

# NIA North had 22 "0 or less than 0 occupancy days" across 10/16/16,
# 10/18/16, 10/28/16-10/30/16, 11/07/16, 11/10/17, and 11/30/16.
# Total: 22 <= 0-values across 8 days.

park[park$Occupancy < 0, ]

# NA's Conclusion: I will only count negative numbers as missing values,
# which there are 12 of.

### Cleaning the initial data set a little and finding means and medians

park_clean <- park[park$SystemCodeNumber != "NIA North", ]
park_clean <- droplevels(park_clean) # So NIA North ID's are dropped from the
# park_clean$SystemCodeNumber factor.

dim(park_clean)
head(park_clean)
park_clean$Percent <- park_clean$Occupancy/park_clean$Capacity

x <- with(park_clean, tapply(Percent, SystemCodeNumber, mean))
park_percent_summary <- data.frame(ID = names(x), mean = x, row.names = 1:29)
# Make data frame of Parking ID's in column 1 and average 
# percent capacity in 
# column 2. Row names set to number, instead of duplicate of Parking ID's.
head(x)

park_percent_summary$sd <- round(with(park_clean,
                                      tapply(Percent, SystemCodeNumber, sd)),2)
park_percent_summary$max <- round(with(park_clean,
                                       tapply(Percent, SystemCodeNumber, max)),2)
park_percent_summary$min <- round(with(park_clean,
                                       tapply(Percent, SystemCodeNumber, min)),2)
park_percent_summary$median <-round(with(park_clean,
                                         tapply(Percent, SystemCodeNumber, median)),2)
park_percent_summary$length <-round(with(park_clean,
                                         tapply(Percent, SystemCodeNumber, length)),2)

park_percent_summary  # these are the same features we used to obtain the clusters 
      # in Figure 4.3 in the textbook. 


####################################################
# We now apply R's tsfeatures() function to obtain #
# other meaningful features for theses time series.#
# Add those features to the park_percent_summary   # 
# data set.Then do a kmeans cluster analysis and   #
# color plot of the clusters. We will compare the  #
# clustering done with the features from tsfeatures# 
# and the clustering obtained from just the        # 
# features already in park_percent_summary         #
####################################################
# install.packages("ggplot2")
library(ggplot2)
#install.packages("tsfeatures")
library(tsfeatures)
#install.packages("tidyverse")
library(tidyverse)
#install.packages("lubridate")
library(lubridate)
#install.packages("dplyr")
library(dplyr)
#install.packages("cluster")
library(cluster)
#install.packages("ggpubr")
library(ggpubr)
#install.packages("factoextra")
library(factoextra)
#install.packages("ggrepel")
library(ggrepel)
#install.packages("GGally")
library(GGally)
# there are 29 unique time series
park_clean %>% 
  group_by(SystemCodeNumber) %>% 
  select(SystemCodeNumber) %>% 
  unique() %>% 
  dim()

# most have more than 1000 observations, not BHMBRTARC01
park_clean %>%
  group_by(SystemCodeNumber) %>% 
  select(SystemCodeNumber, LastUpdated) %>% 
  summarise(
    "Observations" = length(LastUpdated)
  ) %>%
  arrange(Observations) %>% 
  print(n=30)

# We need to remove this one, as it gets grouped alone since it has such
# less data and looks different
park_clean <- park_clean %>% filter(SystemCodeNumber != "BHMBRTARC01")
park_percent_summary <- park_percent_summary %>% filter(ID!= "BHMBRTARC01")

park_clean$LastUpdated <- lubridate::ymd_hms(park_clean$LastUpdated)
park_clean$rounddate <- round_date(park_clean$LastUpdated, unit = "30 minute")

# Plot the data using ggplot2
library(ggrepel)
set.seed(100)
subset <- sample(park_clean$SystemCodeNumber, 6)
plotdata <- park_clean %>% filter(SystemCodeNumber %in% subset) %>% mutate(SystemCodeNumber=as.factor(SystemCodeNumber),rounddate = as.Date(rounddate))
last_points <- plotdata %>% group_by(SystemCodeNumber) %>% slice_tail(n = 1)

ggplot(plotdata, aes(x = rounddate, y = Occupancy, color = SystemCodeNumber)) +
  geom_line() + 
  geom_text_repel(aes(label = SystemCodeNumber, x = max(rounddate)), data = last_points, size = 3, nudge_x = 3) + 
  labs(title = "Occupancy Overtime for 6 Random Lots",
       x = "Time",
       y = "Occupancy") +
  scale_x_date(date_breaks = "1 week", date_labels = "%Y-%m-%d") +
  theme(axis.text.x = element_text(angle = 90, hjust = 1.5),
        legend.position = "none",
        panel.background = element_rect(fill = "white"),
        axis.line = element_line(color = "black")) 
  


### Adding Features to park_percent_summary
tfeatures_data <- park_clean %>% 
  select(SystemCodeNumber, LastUpdated, Percent) %>% 
  group_by(SystemCodeNumber) %>% 
  summarise(
    trend = stl_features(ts(Percent, LastUpdated))[3],
    spike = stl_features(ts(Percent, LastUpdated))[4],
    linearity = stl_features(ts(Percent, LastUpdated))[5],
    acf1 = acf_features(ts(Percent, LastUpdated))[1],
    entropy = entropy(ts(Percent, LastUpdated)),
    hurst= hurst(ts(Percent, LastUpdated)),
    crossing_points = crossing_points(ts(Percent, LastUpdated)),
    flat_spots = flat_spots(ts(Percent, LastUpdated)),
    histogram = histogram_mode(ts(Percent, LastUpdated)))

park_percent_summary <- cbind(park_percent_summary,tfeatures_data[,-1]) 
# combining features into park_percent_summary
head(park_percent_summary);tail(park_percent_summary) # looks good


##########    Performing kmeans()    # ##############
#####################################################

# Select the columns containing tsfeature() features for clustering
features <- park_percent_summary[, 8:15]
head(features) 

# Function to calculate the MSE for a given k
calculate_mse <- function(k) {
  kmeans_result <- kmeans(features, centers = k, nstart = 10)  # Perform k-means clustering
  mse <- sum(kmeans_result$withinss) / nrow(features)  # Calculate MSE
  return(mse)
}

# Determine the range of k values
k_max <- 10

# Calculate MSE for different k values
mse_values<- rep(NA, k_max)
for(i in seq(k_max)){
  mse_values[i]<-calculate_mse(i)
}

# Create a data frame with k and MSE values
elbow_data <- data.frame(K = c(1:k_max), MSE = mse_values)

# Create an elbow plot
ggplot(elbow_data, aes(x = K, y = MSE)) +
  geom_line() +
  geom_point() +
  labs(x = "Number of clusters (k)", y = "Mean Squared Error (MSE)") +
  ggtitle("Elbow Plot for K-means Clustering")

# Perform k-means clustering with k = 3
k <- 3
result <- kmeans(features, centers = k)

# Add cluster assignments and names to the original dataframe
features$clusters <- factor(result$cluster)
park_percent_summary$clusters <- factor(result$cluster)

# Pairs to see separation
library(GGally)
ggpairs(features, mapping= aes(color=clusters))

# Plot the clusters with labels
ggplot(park_percent_summary, aes(x = crossing_points, y = flat_spots, color = clusters, label = ID)) +
  geom_point(size=3) +
  geom_text(size=2, vjust = -0.6) +
  labs(x = "crossing_points", y = "flat_spots", title = "K-means Clustering k=3") +
  scale_color_discrete(name = "Cluster")
